library EADLLWindow;



uses
  SysUtils,
  Classes,
  forms,
  windows,
  messages,
  frmEAU in 'frmEAU.pas' {frmEA} ;



/// Our thread that will be used to provide a message Queue for our Window(s)
type Tdllthread = class(tthread)
protected
   procedure execute; override;
end;




// Global Vars
var
   dllsaveexit : pointer;  // save the pointer to the old dll exit routine we will
                           //
   dllthread : tdllthread; // pointer to our dll thread object

   frmeahandle : hwnd;     // Global reference for our main DLL form.
                           // This variable is used to indicate whether our main form has been created or not
                           // See initDLL and ddlthread.execute



/// create a top most window
/// making sure MT4 is the owner of the window (so that we can make it stay on top of MT4)
procedure createtopwindow(classtype : TComponentClass;var ref);
var
  Instance: TComponent;
  l1 : longint;
begin
  Instance := TComponent(classtype.NewInstance);

  application.handle := mt4apphandle;

  Instance.Create(application);
  TComponent(Ref) := Instance;

  application.handle := DLLapphandle;

  // set the windows properties so that it stays on top
  l1 := getwindowlong(tform(ref).handle,gwl_exstyle);
  l1 := l1 or ws_ex_topmost and not ws_ex_appwindow;
  setwindowlong(tform(ref).handle,gwl_exstyle,l1);
end;






// execute procedure for our new thread with message queue
procedure tdllthread.execute();
var waitresult : dword;
begin



     // we must create first window within our new thread else it wont receive any messages
     createtopwindow(tfrmea,frmea);

     frmea.Show; // we must only show the window
                 // if we showmodal then we will never be able to processmessages
                 // therefore our window will not respond to user input and MT4 will soon crash
                 // If you only want to show modal windows then just hide this base window and
                 // create other modal windows from within the invisible window.

     frmeahandle := frmea.Handle; // copy the handle of our new window into a global var to indicate that this form actually exists
                                  // see dllinit

     // Keep the thread execution alive and enable message processing
     while not self.terminated do
     begin
           application.ProcessMessages; //process all outstanding messages

           // enter idle state until either a std windows message arrives
           // or one of our events is set.
           // In this example only one event is used to wake the thread when the DLL wants to close
           // we could use many more events if required to signal other requirements eg open another window etc.
           // in reality since the user has no direct input method into the EA Code it is unlikely we really need these events

           waitresult := msgwaitformultipleobjects(2,wakeEvents,false,infinite,qs_allinput);
           // if waitresult = 0 then our wakeevent[0] was set because we want to terminate the thread

           //we could use other events to signal other actions if required
           if (waitresult = 1)  then
           begin
                sysutils.beep;
           end;

     end;



     /// kill our form and set its refence to nil
     frmea.free;
     frmea := nil;

     // any other windows we have created should be removed here also


end;





// exported DLL function for our EAs to call
function initdll(pcin :pchar; chartwin : hwnd) : longint; stdcall;
var tmps : String;
    x : integer;
begin
     ///called only from one of out EAS

     entercriticalsection(threadlock); // lock out other threads - both other EAs and our DLL
                                       // because we are accessing the ActiveEAs object and we dont want anyone else to get a chance
                                       // to alter or use it till we are done here


     // If this EAs window is not already active in this DLL then
     if getEAindex(chartwin) =-1 then
     begin

          //If this is the first Time initdll has been called then our DLL Thread will not be running
          // So create it
          if dllthread = nil then
          begin

                // get the window handle of the MT4 Applications main window
                /// we will use this when creating windows so that they belong to the MT4 app rather than our DLL
                mt4apphandle := getparent(chartwin);
                while (getparent(mt4apphandle) <> 0) do mt4apphandle := getparent(mt4apphandle);

                //start our new DLL thread in a running state
                dllthread := tdllthread.Create(false);
                dllthread.Priority := tplower; // set the thread priority to lower (7) this is the same level as a Mt4 EA thread

           end;

          // save the references to this EA
          // We will store the String the EA sent us and use that for the Tabbed control on the main window
          // We will also store the Hwnd of the calling EA to act as a unique identifier for this EA

          tmps := string(pcin);   // save the string (pchar) the EA sent us and store it in a local Delphi string
          setlength(tmps,strlen(pcin)); //make sure the string is the correct length
          x := activeEas.addobject(tmps,tobject(chartwin)); // for now use the Object property in the strings list to store the new Charthwnd




          // we must wait here until our main form has been created by our DLLThread
          // Before we can send a message to our form
          while frmeahandle = 0 do
          begin
               sleep(5);
          end;


          // Sendmessage to the main window to make it add a new tab for this EA
          // x = the index in activeEas string list object of the new EA to add
          // WE use SEndmessage because WE want to wait until the form has created
          // our new tab and controls so we rule out the unlikely event that the EA gets to request another item
          // before the completion of the creation of our objects.
          sendmessage(frmea.handle,custommsgval,msg_opennewtab,x);


          result := 1;
     end
     else result := -1; // chart window is already active in this DLL

     leavecriticalsection(threadlock); // allow other threads (EAs)

end;


// final clean up code executed when all attached EAs have called deinit
Procedure killDLLThread;
begin

     if dllthread <> nil then
     begin


          // if the modal dialog form is open then close it
          if frmmodaldialog <> nil then postmessage(frmmodaldialog.handle,wm_close,0,0);


          // kill the DLL thread and wait for it to terminate
          dllthread.Terminate; // signal we want the thread to terminate
          setevent(wakeEvents[0]); // wake the thread up incase it was in an idle state
          dllthread.WaitFor; // wait for thread to terminate
          dllthread.Free; // kill the thread object
          dllthread := nil; // set the reference to nil because we use this to check if the thread is already created

          frmmodaldialog := nil;      // set our global reference to any open Dialogs to nil
          frmeahandle :=0;            // set our global reference to our main DLL form to 0 - see var declaration for info

     end;

end;


// exported DLL function for our EAs to call
function deinitdll(charthwnd : hwnd) : longint; stdcall;
var x: longint;
begin
     entercriticalsection(threadlock); // lock out other threads   (EAs)

     if dllthread <> nil then
     begin
          // if we find this charts Hwnd then delete it from our ActiveEAs list
          /// and also inform the main window that we want to remove the tab for this EA
          x := getEaindex(charthwnd);
          if x <> -1 then
          begin
               /// delete references for this EA
               activeeas.Objects[x].free; // free our EAobj for this EA
               activeEAs.delete(x);  // delete this EAs entry

               // sendmessage to main window to remove the tab for this EA
               // we want to wait using SendMessage incase our user clicks on the tab control whilst
               // we are half way through removing the EA tab.
               sendmessage(frmea.handle,custommsgval,msg_closetab,x);

          end;

          // If there are no EAs left then kill the DLL
          if activeEas.count = 0 then KilldllThread;

          result := 1;
     end
     else result := -1;

     leavecriticalsection(threadlock); // allow other threads (EAs)
end;





// called by an EA to see if it has any pending incoming messages for our DLL Thread window

// ********************************************************************
// In the getmsg procedure we would ideally like to simply do something like  :
//     result := pchar(messagesout.strings[0]);
//     messagesout.delete(0);
// however Delphi cannot and does not copy the string into the result variable
// only a pointer to the String is passed back to the EA. MT4 then copies the string (behind the scenes) into its own address space
// when the function returns
// If we immediately delete the string from the stringlist Delphi will free the memory for the string
// that our EA looks at. This will sometimes be corrupted since it is no longer a valid memory location
//     IF there was only a single EA using the DLL then we could simply use a global string var to temporarily store
//     the message string for the EA to read ... something like :
//         tmpglobalmsg := messagesout.strings[0];
//         result := pchar(tmpglobalmsg);
//         messagesout.delete(0);
//
//  However in our case we have multiple EA threads calling so we need to be able to keep the string alive until
//  we Know that the EA is done with it.
//  In order to achieve this I have implemented a messagessent stringlist for this temporary storage.
//  Each message is given a unique ID (I've used the current time in milliseconds).
//  Passing this unique value back to the EA could be tricky but I've simply opted to tag the ID onto the end of the message CSV string
//  this unique ID can then be used by our EA to tell this DLL that we can delete the string
//  the EA does this by calling the removemsg DLL procedure after every call to getmsg that returns a message.

function getmsg(chartwin : dword) : pchar; stdcall;
var x,y : longint;
    dt : ttimestamp;
begin
     result := pchar('');
     if Messagesout.Count = 0 then exit;

     x :=0;
     entercriticalsection(threadlock); // lock out other threads   (EAs)

     /// USE FIFO (First in first out)  find if there are any messages for this EA
     // and return the message string if we do find one
     while x < Messagesout.count do
     begin
          if dword(messagesout.objects[x]) = chartwin then
          begin
               dt := datetimetotimestamp(now);
               // save the outgoing message in the messagesent list and tag the time in millisecs onto the end
               // to act as A unique message ID
               y := messagessent.AddObject(messagesout.strings[x]+','+inttostr(dt.time),tobject(dt.time));
               // Return the new string in the messagessent object
               result := pchar(messagessent.strings[y]);
               // delete the out going message from the messagesout object
               messagesout.delete(x);
               // quit while loop
               break;
          end;
          inc(x);
     end;

     leavecriticalsection(threadlock); // allow other threads (EAs)
end;


// exported DLL function for our EAs to call
// used to remove a message from the messagessent string list now the EA is done with it
procedure removesentmsg(msgtime : dword); stdcall;
var x: longint;
begin
     entercriticalsection(threadlock); // lock out other threads   (EAs)

     x := messagessent.IndexOfObject(tobject(msgtime));
     if x <> -1 then messagessent.delete(x);

     leavecriticalsection(threadlock); // allow other threads (EAs)

end;



/// called by an EA to post a message to out DLL thread Window requesting an action of some sort
procedure postmsg(chartwin : dword;pcin : pchar); stdcall;
var tmpstr : string;
begin

     tmpstr := string(pcin);
     setlength(tmpstr,strlen(pcin));

     entercriticalsection(threadlock); // lock out other threads   (EAs)

     // only lock while we are actually changing the stringlist object
     // save the message string and the EAs Charthwnd so we know which EA is requesting this action or send this data
     Messagesin.AddObject(tmpstr,tobject(chartwin));

     leavecriticalsection(threadlock); // allow other threads (EAs)

     ///  postmessage to signal our window that a new message has been added
     // Use postmessage because we really dont need or want to wait for our form to process it!
     postmessage(frmea.handle,custommsgval,msg_processmsglist,0);

end;


// Final exit code for the DLL only called when the DLL is unloaded from memory
// MT4 does not forceably unload a DLL for example If you recompile the MQ4 file when the EA is running
// therefore This code will not execute until all EAs are removed from the chart or the MT4 app is Closed.
procedure LibExit;
var x : longint;
begin
     killDllThread;  // if the DLL is shutting down make sure all our forms and objects are destroyed
                      // incase the user did not call Deinit themselves


      // remove our events for thread wakeup
      for x := 0 to maxevents-1 do Closehandle(wakeEvents[x]);

      // kill our string list objects
      ActiveEAs.free;
      messagesIn.free;
      messagesOut.free;
      messagessent.free;

     DeleteCriticalSection(Threadlock); // delete our thread lock object

     ExitProc := dllsaveexit;  // restore exit procedure chain
end;


//// DLL Function exports
exports    initdll index 1,
           deinitdll index 2,
           getmsg index 3,
           postmsg index 4,
           removesentmsg index 5;



/// Executes only when the DLL is first loaded.
begin
     /// some initialisation when the DLL is first loaded.
     custommsgval := registerwindowmessage(pchar('MT4EADLLcustomMessage')); // create a custom windows message value for later use - to keep things thread safe

     DLLapphandle := application.handle; // save the application handle of this DLL


     InitializeCriticalSection(Threadlock);  /// create a lock obect so we can lock out other threads



     activeEAs := tstringlist.create;   //used to monitor all the separate EAs that are using this DLL
     messagesIn := tstringlist.create;  // Used to store incoming messages
     messagesOut := tstringlist.create; // Used To store outgoing messages
     messagessent := tstringlist.create; // Used To store outgoing messages that have already been sent

     // create some events so we can signal our thread directly
     // dont need to give each event a name eg 'wakedllthread' .. we could use NIL for all of them
     wakeEvents[0] := createevent(NIL,false,false,pchar('WakeDllThread'));
     wakeEvents[1] := createevent(NIL,false,false,pchar('Beep'));


     mt4apphandle := 0;
     frmEA := nil;
     dllthread := nil;
     frmmodaldialog := nil;  // set all object pointers to  nil ... probably overkill but !!



     dllsaveexit := Exitproc;
     ExitProc := @LibExit; // the function libexit will now be called should the DLL be unloaded before ,
                           // deinitdllform is called by all EAs that are using this DLL
end.
