unit frmEAU;

/// This DLL & Window are designed to interface with multiple owning threads (EAs)
/// because of this (and to aid quick development) I have implemented dynamically created
/// window/visual components (namely the panel,labels,button, and Memo objects) .
/// This leaves the programmer without the ability to set these components attributes /properties at design time.
/// This could be remedied by using a single panel/ memo etc setting them up at design time
/// and then use other objects (tstringslist or the memo etc) to to store the values for each calling EA
/// and only display the relevant values according to the Tab controls selected item.


interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, Buttons, ComCtrls, ExtCtrls;


// forward declared function used in the DLL Unit
function getEAindex(chartwnd : dword) : longint;



type
  TfrmEA = class(TForm)
    Label1: TLabel;
    btn2: TButton;
    Button1: TButton;
    Button2: TButton;
    TabControl1: TTabControl;
    procedure btn2Click(Sender: TObject);
    procedure Button1Click(Sender: TObject);
    function showdialogontop(textstr : string) : longint;
    procedure Button2Click(Sender: TObject);
    procedure TabControl1Change(Sender: TObject);
    procedure buttonclicks(Sender: TObject);
    procedure makenewtabforEA(EAindex : longint);
    procedure processEAmessage(msg: String;charthw : dword);
    procedure editexit(Sender: TObject);
  private
    { Private declarations }
  public
    { Public declarations }
  protected
    procedure wndproc(var mmessage :tmessage); override;  // override this windows wndproc so we can catch messages for our custommsgval
  end;



  //// small object to hold a few vars and references for each of our EAs
  tEAobj = Class(tobject)
  public
      charthwnd : dword;
      EAlabel : tlabel;
      EAMemo : tmemo;
      EAedit : tedit;
  end;




const
     maxevents =2;

     msg_opennewtab = 1;           // used in messaeg.wparam to signify different actions for our custom windows message : custommsgval
     msg_processmsglist = 2;
     msg_closetab =3;


/// Global form Vars
/// Since we cant have circular uses references Some vars
/// that would be better in the DLL unit now reside here
/// Probably better in a globals Unit
var
  frmEA: TfrmEA;

  mt4apphandle,DLLapphandle : hwnd; // stores the Application handles for both MT4 and this DLL.
                                    // These are used when creating windows so we can ensure MT4 owns them
                                    // In this way our windows can be made to stay on top of MT4

  custommsgval : dword;   // stores the value of a custom windows message that we will use to tell our form it needs to do something for us.

  wakeEvents : array[0..maxevents-1] of dword;  // used to signal our new thread

  frmmodaldialog : tform;   // global pointer to our modal dialog window

  ThreadLock: TRTLCriticalSection;  // Used to force other threads to wait until there are no other threads using the either this section of code.
                                    // or the objects that are thread sensitive

                                    // In most of these following objects we use the Object property to store
                                    // a longint ... usually the unique chart hwnd so we know which string (entery) belongs to which EA
  messagesIn,                       // text based message queue for incoming messages (from EAs)
  messagesOut,                      // text based message queue for outgoing messages (to EAs)
  messagessent,                     // Stores messages that have already been sent **** SEE function getmsg for more info
  ActiveEAs : tstringlist;          // List used to store the current EAs that are active or using this DLL



implementation




{$R *.DFM}


/// utility function returning the requested field (position) in a comma seperated string
function getcsvval(ts:string;posi:integer):string;
var x,spos : integer;
    rstr :string;
begin
     x:= 1;
     spos :=1;
     rstr := '';
     result := '';
     if ts = '' then exit;
     while (x < posi) do
     begin
          if copy(ts,spos,1)[1] = ',' then inc(x);
          inc(spos);
     end;

     while (spos <= length(ts)) and (copy(ts,spos,1)[1] <> ',') do
     begin
          rstr := rstr+copy(ts,spos,1);
          inc(spos);
     end;

     result := rstr;

end;


/// get the EA index value in the activeEAs stringlist object using the Charts unique Hwnd
function getEAindex(chartwnd : dword) : longint;
begin
     result := ActiveEAs.Count-1;
     if result < 0 then exit;

     while (result >=0) do
     begin
          if tEAobj(ActiveEas.Objects[result]).charthwnd = chartwnd then break
          else dec(result);
     end;

end;





function TfrmEA.showdialogontop(textstr : string) : longint;
begin
     // This is the safe way to show a Modal window ... we are using global variable to reference the window
     // so that if all EAs are removed when this window is open we can close it in the deinit process.
     // The Dialog window is only modal to our main form
     // We can still interact with MT4 but we cant interact with our main form
     // ***** NOTE this window's pointer is stored globally This is because MT4 may unload this DLL when this window is open
     // therefore we need to be able to close and destroy it as we execute the DLL exit code


     // create a new mesage dialog form using MT4's application handle
     application.handle := mt4apphandle;
     frmmodaldialog := CreateMessageDialog(textstr , mtcustom, [mbok,mbcancel]);
     application.handle := DLLapphandle;


     /// Show the Dialog window in a modal manner and return the result.
     with frmmodaldialog do
     try
      HelpContext := 0;
      HelpFile := '';

      // set window to topmost
      setwindowpos(handle,hwnd_topmost or hwnd_top,self.left,self.top,0,0,swp_nosize);
      result:= ShowModal;
     finally
      Free;
      frmmodaldialog := nil; /// reset our global refernce for this window -used to in killDLLthread to close this window if it was already open
     end;
end;


procedure TfrmEA.btn2Click(Sender: TObject);
begin
     // button event for unsafe dialog window

     // this is a bad way of showing a modal message dialog window.

     showmessagepos('modal dialog example. This could be our own custom dialog form.'+#13#10+'NOT Modal to the MT4 application ONLY to our EA window IE you can interact with MT4 but not with our other EA window'+#13#10+'Note this window is below our Topmost Indicator window. and will fall behind MT4 if we click the MT4 App'+#13#10+
     'Since MT4 still responds we can remove the EA from the window in the normal way. Removing the EA will not always complete if this modal window is open, and MT4 will often crash soon after'+#13#10+'Even if the EA appears to remove correctly it is possible for MT4 to crash tens of minutes later!',self.left+20,self.top+20);
end;


procedure TfrmEA.Button1Click(Sender: TObject);
var res :longint;
begin
     // button event for SAFE dialog window

     res := showdialogontop('Modal Message dialog window ... This has been set to the Topmost Window'+#13#10+'Code has been implemented to safely close this dialog If you remove the EA whilst this window is open and therefore MT4 will not crash');
     if res = mrok then label1.caption := 'result= OK' else label1.caption := 'result= Cancel';
end;



// indirectly called by the EA on a call to InitDLL
// This proc setups all visual controls and saves all required data for each EA
procedure tfrmea.makenewtabforEA(EAindex : longint);
var
    panel : tpanel;
    lbl : tlabel;
    btn : tbutton;
    memo : tmemo;
    edit : tedit;
    tmpobj : tEAobj;
    x : longint;
begin
     // NB this proc should ONLY be called from wndproc as
     // our threadlocks have already been implemented.


        // create a new EAobj to store control references.
       tmpobj := teaobj.create;
       // save the EAs Hwnd in the EAobj
       tmpobj.charthwnd := dword(activeEas.objects[EAindex]);

       // Create a new panel for this tab
       panel := tpanel.create(self);
       panel.Parent := twincontrol(frmea);
       panel.left :=0;
       panel.top := tabcontrol1.Height;
       panel.height := frmea.Height -120;
       panel.width := frmea.width-15;
       panel.visible := false;
       panel.Caption :=activeEas.strings[EAindex];

       // Create a new label on the panel
       lbl := tlabel.create(self);
       lbl.Parent := panel;
       lbl.Top := 10;
       lbl.left := 10;
       lbl.Height := 30;
       lbl.Caption := 'Current Price';
       lbl.Width := 100;

       // Create another new label on the panel
       lbl := tlabel.create(self);
       lbl.Parent := panel;
       lbl.Top := 30;
       lbl.left := 10;
       lbl.Height := 30;
       lbl.Caption :='';
       lbl.Width := 100;

       // save refernce to label
       tmpobj.EAlabel := lbl;

       // create a new button
       btn := tbutton.Create(self);
       btn.parent := panel;
       btn.Top := 70;
       btn.Left := 10;
       btn.Height := 30;
       btn.width := 110;
       btn.caption := 'Send RSI Request';
       btn.tag := longint(activeEas.objects[EAindex]);
       btn.OnClick := buttonclicks;

       // create a new edit control
       edit := tedit.Create(self);
       edit.parent := panel;
       edit.top := 110;
       edit.left := 10;
       edit.width := 110;
       edit.Height := 25;
       edit.text := '14';
       edit.OnExit := editexit;

       //save a reference to theis Edit control
       tmpobj.EAedit := edit;

       // create a memo
       memo := tmemo.create(self);
       memo.parent := panel;
       memo.Top := 20;
       memo.Left := 130;
       memo.Width := 300;
       memo.Height := 250;
       memo.ScrollBars := ssvertical;
       memo.Tag := longint(activeEas.objects[EAindex]);

       //save reference to memo
       tmpobj.eamemo := memo;

       // save reference for our EAobj in the string list.
       ActiveEas.objects[EAindex] := tmpobj;

       // create a new tab and save the reference for its associated panel
       // so we can show it when the user changes tabs.
       x := tabcontrol1.tabs.Addobject(activeEas.strings[EAindex],panel);
       tabcontrol1.TabIndex := x;
       TabControl1Change(nil);
end;


/// This proc will process all the text (csv) based messages we receive from all the EAs
procedure tfrmea.processEAmessage(msg: String;charthw : dword);
var x: longint;
    tstr : string;
begin
     // you want to order this If statement (or case statement) (if possible) to process the most frequent messages first
     tstr := getcsvval(msg,1);

     if tstr = 'CURPRICE' then
     begin
          x := getEAindex(charthw);
          tEAobj(ActiveEas.Objects[x]).EAlabel.caption := getcsvval(msg,2);
     end
     else
     if tstr = 'LOGMSG' then
     begin
          x := getEAindex(charthw);
          tEAobj(ActiveEas.Objects[x]).EAmemo.lines.Add(getcsvval(msg,2));
     end
     else
     if tstr = 'RSIRESULT' then
     begin
          x :=getEAindex(charthw);
          tEAobj(ActiveEas.Objects[x]).EAmemo.lines.Add(getcsvval(msg,2));
     end;

end;


// our overriden wndproc for our window
// this is overriden so we can catch our custom messageval messages
procedure tfrmea.wndproc(var mmessage :tmessage);
begin
     if mmessage.msg = custommsgval then
     begin
          if mmessage.WParam = msg_opennewtab then
          begin
               // mmessage.lparam  = the index within the activeEAs list object for this newly attached EA

               // add a new tab for this EA

               // we have no need to use threadlock here since this message is only called
               // using Sendmessage which will wait until we have finished here
               // the EA calling our DLL has already locked the threadlock so we dont need to
               // in fact is we try to here our DLL will crash.

               makenewtabforEA(mmessage.lparam);
          end
          else
          if mmessage.wparam = msg_closetab then
          begin
               /// mmessage.lparam = the index of this EAs entry in the  ActiveEAs list object.

               // delete the panel and tab for this EA that has just Deinited.
               tpanel(tabcontrol1.Tabs.Objects[mmessage.lparam]).free;
               // remove the tab from the tab control
               tabcontrol1.Tabs.Delete(mmessage.lparam);
               tabcontrol1.TabIndex := 0;
               TabControl1Change(nil);
          end
          else
          if mmessage.WParam = msg_processmsglist then
          begin
               entercriticalsection(threadlock); // lock out other threads   (EAs)

               // Process all new incoming messages
               while Messagesin.count > 0 do
               begin
                    processEAmessage(Messagesin.strings[0],dword(Messagesin.objects[0]));
                    Messagesin.Delete(0);
               end;

               leavecriticalsection(threadlock); // allow other threads (EAs)
          end;
     end
     else
     // if its not a message for us then make sure we pass it on to the message handler we overrode
     inherited wndproc(mmessage);
end;



/// button click event for all our EA buttons - 'Send RSI Request'
procedure TfrmEA.buttonclicks(Sender: TObject);
var x : longint;
begin


     entercriticalsection(threadlock); // lock out other threads   (EAs)
     // add a CSV string to the outbound messages list and also save the unique (hwnd) for the target EA
     // so we know which message to belongs to which EA (see getmsg function in EADLLWINDOW)

     x := geteaindex(tbutton(sender).tag);
     if x <> -1 then
     begin

          messagesout.AddObject('GETRSI,'+teaobj(activeeas.Objects[x]).eaedit.text,tobject(tbutton(sender).tag));
     end;

     leavecriticalsection(threadlock); // allow other threads (EAs)
end;

/// onexit event for all our EA edit controls
procedure TfrmEA.editexit(Sender: TObject);
begin

     /// test edit box entry to make sure its a numeric string

     try
        strtoint(tedit(sender).text);
     except
        tedit(sender).text := '';
     end;

     // if not then reset it and warn user
     if tedit(sender).text = '' then
     begin
          beep;
          label1.caption := 'NOT NUMERIC';
          tedit(sender).text := '14';
     end;
end;

// Button click event for doing a lengthy calculation / operation
procedure TfrmEA.Button2Click(Sender: TObject);
var
  iter,t1 : dword;
begin
   t1 :=0;
  iter := $8FFF;
  while iter >0 do
  begin
       t1 := 1000*t1;
       t1 := t1 div 2;
       dec(iter);
       label1.caption := inttostr(iter);
       label1.refresh;

       application.processmessages;
  end;

end;

// Change event for the Tab control so we can bring the correct panel for this EA into view.
procedure TfrmEA.TabControl1Change(Sender: TObject);
var x : longint;
begin
     if tabcontrol1.Tabs.Count > 0 then
     begin
          // hide all panels
          for x := 0 to tabcontrol1.tabs.count-1 do
          tpanel(tabcontrol1.Tabs.Objects[x]).visible := false;

          // show the panel for the selected EA
          tpanel(tabcontrol1.Tabs.Objects[tabcontrol1.TabIndex]).visible := true;
     end;

end;

end.
